// calculate integral using full-workers model
// compile:  mpicc integral3_mpi.c -lm
// usage: mpirun -n process# ./a.out trial#

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <mpi.h>


// w is gaussian weight function
double w(double x, double y)
{
  return exp( - ( x*x + y*y ) );

}

//f is the function to be integrated
double f(double x, double y)
{

   return M_PI *  ( x*x + y*y ) ;

}


void generator(double * x, double * y)
{
   double xt, yt, ratio, tmp ;
   double delta = 0.5 ;


   tmp = (double)rand() / (double)RAND_MAX  ;
   xt = (*x) + delta * (2 * tmp - 1) ; 

   tmp = (double)rand() / (double)RAND_MAX  ;
   yt = (*y) + delta * (2 * tmp - 1) ; 

   ratio = w(xt, yt ) / w(*x, *y) ;
   tmp = (double)rand() / (double)RAND_MAX  ;

   if(ratio > tmp )
   {

      *x = xt ;
      *y = yt ;
   }


}


int main(int argc, char* argv[])
{
   double  x, y ;
   int generated = 0;
   double sum = 0;
   double  total_sum ;
   double integral ;
   int trials, total_trial;
   int i, j ;
   int pnumber, myrank;
   int my_tag1 = 1111 ;
   int my_tag2 = 2222 ;


   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
   MPI_Comm_size(MPI_COMM_WORLD, &pnumber);




   x = 0.0 ;
   y = 0.0 ;


   trials = atoi(argv[1]);


   MPI_Comm grid_comm;
   int dim_sizes[2];
   int periods[2];
   int reorder = 1;
   dim_sizes[0] = sqrt(pnumber);
   dim_sizes[1] = sqrt(pnumber);
   periods[0] = 0; periods[1] = 0;
   int maxdims = 2 ;
   int my_grid_rank;
   MPI_Cart_create(MPI_COMM_WORLD, maxdims, dim_sizes, periods, reorder, &grid_comm); 
   MPI_Comm_rank(grid_comm, &my_grid_rank);
   int coordinates[2];
   MPI_Cart_coords(grid_comm, my_grid_rank, maxdims, coordinates);





   for(i=0 ;  i < trials ; i++)
   {
      generator(&x, &y);
      sum += f(x,y);
      generated++ ;
   }



   MPI_Reduce(&sum, &total_sum, 1, MPI_DOUBLE , MPI_SUM, 0, MPI_COMM_WORLD) ;
   MPI_Reduce(&generated, &total_trial, 1, MPI_INT , MPI_SUM, 0, MPI_COMM_WORLD) ;





   if(myrank == 0)
   {

      printf("%f\n", total_sum / total_trial);

   }

   MPI_Finalize();


   return 0;

}
